@@ -31,8 +31,8 @@ static int default_layout(struct supertype *st, int level, int verbose)
 {
 	int layout = UnSet;
 
-	if (st && st->ss->default_layout)
-		layout = st->ss->default_layout(level);
+	if (st && st->ss->default_geometry)
+		st->ss->default_geometry(st, &level, &layout, NULL);
 
 	if (layout == UnSet)
 		switch(level) {
@@ -43,7 +43,7 @@ static int default_layout(struct supertype *st, int level, int verbose)
 			layout = 0x102; /* near=2, far=1 */
 			if (verbose > 0)
 				fprintf(stderr,
-					Name ": layout defaults to n1\n");
+					Name ": layout defaults to n2\n");
 			break;
 		case 5:
 		case 6:
@@ -66,11 +66,13 @@ static int default_layout(struct supertype *st, int level, int verbose)
 
 
 int Create(struct supertype *st, char *mddev,
-	   int chunk, int level, int layout, unsigned long long size, int raiddisks, int sparedisks,
+	   int chunk, int level, int layout, unsigned long long size,
+	   int raiddisks, int sparedisks,
 	   char *name, char *homehost, int *uuid,
-	   int subdevs, mddev_dev_t devlist,
+	   int subdevs, struct mddev_dev *devlist,
 	   int runstop, int verbose, int force, int assume_clean,
-	   char *bitmap_file, int bitmap_chunk, int write_behind, int delay, int autof)
+	   char *bitmap_file, int bitmap_chunk, int write_behind,
+	   int delay, int autof)
 {
 	/*
 	 * Create a new raid array.
@@ -93,7 +95,7 @@ int Create(struct supertype *st, char *mddev,
 	char *mindisc = NULL;
 	char *maxdisc = NULL;
 	int dnum;
-	mddev_dev_t dv;
+	struct mddev_dev *dv;
 	int fail=0, warn=0;
 	struct stat stb;
 	int first_missing = subdevs * 2;
@@ -120,15 +122,8 @@ int Create(struct supertype *st, char *mddev,
 	int major_num = BITMAP_MAJOR_HI;
 
 	memset(&info, 0, sizeof(info));
-
-	if (level == UnSet) {
-		/* "ddf" and "imsm" metadata only supports one level - should possibly
-		 * push this into metadata handler??
-		 */
-		if (st && (st->ss == &super_ddf || st->ss == &super_imsm))
-			level = LEVEL_CONTAINER;
-	}
-
+	if (level == UnSet && st && st->ss->default_geometry)
+		st->ss->default_geometry(st, &level, NULL, NULL);
 	if (level == UnSet) {
 		fprintf(stderr,
 			Name ": a RAID level is needed to create an array.\n");
@@ -169,15 +164,15 @@ int Create(struct supertype *st, char *mddev,
 		    inf.raid_disks == 0) {
 			/* yep, looks like a container */
 			if (st) {
-				rv = st->ss->load_super(st, fd,
-							devlist->devname);
+				rv = st->ss->load_container(st, fd,
+							    devlist->devname);
 				if (rv == 0)
 					have_container = 1;
 			} else {
-				st = guess_super(fd);
+				st = super_by_fd(fd, NULL);
 				if (st && !(rv = st->ss->
-					    load_super(st, fd,
-						       devlist->devname)))
+					    load_container(st, fd,
+							   devlist->devname)))
 					have_container = 1;
 				else
 					st = NULL;
@@ -235,9 +230,11 @@ int Create(struct supertype *st, char *mddev,
 	case 6:
 	case 0:
 		if (chunk == 0) {
-			chunk = 512;
+			if (st && st->ss->default_geometry)
+				st->ss->default_geometry(st, NULL, NULL, &chunk);
+			chunk = chunk ? : 512;
 			if (verbose > 0)
-				fprintf(stderr, Name ": chunk size defaults to 512K\n");
+				fprintf(stderr, Name ": chunk size defaults to %dK\n", chunk);
 		}
 		break;
 	case LEVEL_LINEAR:
@@ -375,20 +372,37 @@ int Create(struct supertype *st, char *mddev,
 			warn |= check_ext2(fd, dname);
 			warn |= check_reiser(fd, dname);
 			warn |= check_raid(fd, dname);
-			if (st && strcmp(st->ss->name, "1.x") == 0 &&
+			if (strcmp(st->ss->name, "1.x") == 0 &&
+			    st->minor_version >= 1)
+				/* metadata at front */
+				warn |= check_partitions(fd, dname, 0);
+			else if (level == 1 || level == LEVEL_CONTAINER)
+				/* partitions could be meaningful */
+				warn |= check_partitions(fd, dname, freesize*2);
+			else
+				/* partitions cannot be meaningful */
+				warn |= check_partitions(fd, dname, 0);
+			if (strcmp(st->ss->name, "1.x") == 0 &&
 			    st->minor_version >= 1 &&
 			    did_default &&
-			    level == 1) {
-				warn = 1;
+			    level == 1 &&
+			    (warn & 1024) == 0) {
+				warn |= 1024;
 				fprintf(stderr, Name ": Note: this array has metadata at the start and\n"
 					"    may not be suitable as a boot device.  If you plan to\n"
-					"    store '/' or '/boot' on this device please ensure that\n"
+					"    store '/boot' on this device please ensure that\n"
 					"    your boot-loader understands md/v1.x metadata, or use\n"
-					"    --metadata=1.0\n");
+					"    --metadata=0.90\n");
 			}
 			close(fd);
 		}
 	}
+	if (raiddisks + sparedisks > st->max_devs) {
+		fprintf(stderr, Name ": Too many devices:"
+			" %s metadata only supports %d\n",
+			st->ss->name, st->max_devs);
+		return 1;
+	}
 	if (have_container)
 		info.array.working_disks = raiddisks;
 	if (fail) {
@@ -527,7 +541,7 @@ int Create(struct supertype *st, char *mddev,
 	     assume_clean
 		) {
 		info.array.state = 1; /* clean, but one+ drive will be missing*/
-		info.resync_start = ~0ULL;
+		info.resync_start = MaxSector;
 	} else {
 		info.array.state = 0; /* not clean, but no errors */
 		info.resync_start = 0;
@@ -599,7 +613,7 @@ int Create(struct supertype *st, char *mddev,
 
 	total_slots = info.array.nr_disks;
 	sysfs_init(&info, mdfd, 0);
-	st->ss->getinfo_super(st, &info);
+	st->ss->getinfo_super(st, &info, NULL);
 
 	if (did_default && verbose >= 0) {
 		if (is_subarray(info.text_version)) {
@@ -640,6 +654,11 @@ int Create(struct supertype *st, char *mddev,
 			fprintf(stderr, Name ": internal bitmaps not supported by this kernel.\n");
 			goto abort;
 		}
+		if (!st->ss->add_internal_bitmap) {
+			fprintf(stderr, Name ": internal bitmaps not supported with %s metadata\n",
+				st->ss->name);
+			goto abort;
+		}
 		if (!st->ss->add_internal_bitmap(st, &bitmap_chunk,
 						 delay, write_behind,
 						 bitmapsize, 1, major_num)) {
@@ -652,7 +671,7 @@ int Create(struct supertype *st, char *mddev,
 
 	sysfs_init(&info, mdfd, 0);
 
-	if (st->ss->external && st->subarray[0]) {
+	if (st->ss->external && st->container_dev != NoMdDev) {
 		/* member */
 
 		/* When creating a member, we need to be careful
@@ -715,7 +734,7 @@ int Create(struct supertype *st, char *mddev,
 	infos = malloc(sizeof(*infos) * total_slots);
 
 	for (pass=1; pass <=2 ; pass++) {
-		mddev_dev_t moved_disk = NULL; /* the disk that was moved out of the insert point */
+		struct mddev_dev *moved_disk = NULL; /* the disk that was moved out of the insert point */
 
 		for (dnum=0, dv = devlist ; dv ;
 		     dv=(dv->next)?(dv->next):moved_disk, dnum++) {
@@ -755,7 +774,8 @@ int Create(struct supertype *st, char *mddev,
 				if (have_container)
 					fd = -1;
 				else {
-					if (st->ss->external && st->subarray[0])
+					if (st->ss->external &&
+					    st->container_dev != NoMdDev)
 						fd = open(dv->devname, O_RDWR);
 					else
 						fd = open(dv->devname, O_RDWR|O_EXCL);
@@ -773,9 +793,11 @@ int Create(struct supertype *st, char *mddev,
 				if (fd >= 0)
 					remove_partitions(fd);
 				if (st->ss->add_to_super(st, &inf->disk,
-							 fd, dv->devname))
+							 fd, dv->devname)) {
+					ioctl(mdfd, STOP_ARRAY, NULL);
 					goto abort;
-				st->ss->getinfo_super(st, inf);
+				}
+				st->ss->getinfo_super(st, inf, NULL);
 				safe_mode_delay = inf->safe_mode_delay;
 
 				if (have_container && verbose > 0)
@@ -801,7 +823,6 @@ int Create(struct supertype *st, char *mddev,
 						Name ": ADD_NEW_DISK for %s "
 						"failed: %s\n",
 						dv->devname, strerror(errno));
-					st->ss->free_super(st);
 					goto abort;
 				}
 				break;
@@ -820,7 +841,7 @@ int Create(struct supertype *st, char *mddev,
 			 * again returns container info.
 			 */
 			map_lock(&map);
-			st->ss->getinfo_super(st, &info_new);
+			st->ss->getinfo_super(st, &info_new, NULL);
 			if (st->ss->external && level != LEVEL_CONTAINER &&
 			    !same_uuid(info_new.uuid, info.uuid, 0)) {
 				map_update(&map, fd2devnum(mdfd),
@@ -835,7 +856,7 @@ int Create(struct supertype *st, char *mddev,
 			if (me) {
 				char *path = strdup(me->path);
 
-				st->ss->getinfo_super(st, &info_new);
+				st->ss->getinfo_super(st, &info_new, NULL);
 				map_update(&map, st->container_dev,
 					   info_new.text_version,
 					   info_new.uuid, path);
@@ -844,10 +865,10 @@ int Create(struct supertype *st, char *mddev,
 			map_unlock(&map);
 
 			flush_metadata_updates(st);
+			st->ss->free_super(st);
 		}
 	}
 	free(infos);
-	st->ss->free_super(st);
 
 	if (level == LEVEL_CONTAINER) {
 		/* No need to start.  But we should signal udev to
@@ -858,33 +879,40 @@ int Create(struct supertype *st, char *mddev,
 		wait_for(chosen_name, mdfd);
 	} else if (runstop == 1 || subdevs >= raiddisks) {
 		if (st->ss->external) {
+			int err;
 			switch(level) {
 			case LEVEL_LINEAR:
 			case LEVEL_MULTIPATH:
 			case 0:
-				sysfs_set_str(&info, NULL, "array_state",
-					      "active");
+				err = sysfs_set_str(&info, NULL, "array_state",
+						    "active");
 				need_mdmon = 0;
 				break;
 			default:
-				sysfs_set_str(&info, NULL, "array_state",
-					      "readonly");
+				err = sysfs_set_str(&info, NULL, "array_state",
+						    "readonly");
 				break;
 			}
 			sysfs_set_safemode(&info, safe_mode_delay);
+			if (err) {
+				fprintf(stderr, Name ": failed to"
+					" activate array.\n");
+				ioctl(mdfd, STOP_ARRAY, NULL);
+				goto abort;
+			}
 		} else {
 			/* param is not actually used */
 			mdu_param_t param;
 			if (ioctl(mdfd, RUN_ARRAY, &param)) {
 				fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 					strerror(errno));
-				Manage_runstop(mddev, mdfd, -1, 0);
+				ioctl(mdfd, STOP_ARRAY, NULL);
 				goto abort;
 			}
 		}
 		if (verbose >= 0)
 			fprintf(stderr, Name ": array %s started.\n", mddev);
-		if (st->ss->external && st->subarray[0]) {
+		if (st->ss->external && st->container_dev != NoMdDev) {
 			if (need_mdmon)
 				start_mdmon(st->container_dev);
 
@@ -899,6 +927,10 @@ int Create(struct supertype *st, char *mddev,
 	return 0;
 
  abort:
+	map_lock(&map);
+	map_remove(&map, fd2devnum(mdfd));
+	map_unlock(&map);
+
 	if (mdfd >= 0)
 		close(mdfd);
 	return 1;
